#!/usr/bin/perl -w
#####################################################################################################################################
#
#cdBEST - cdBEST is a tool/software developed in Perl to search the chromatin domain boundary element sequences in
#12 Drosophila species. cdBEST uses recognition sequences of boundary interacting proteins and looks for motif clusters
#under a set of constrains to predict boundary element sequences. Apart from Drosophila, cdBEST can also be used in
#several other insect genomes, where the boundary interacting proteins are conserved.
#
#Copyright (C) 2011  Arumugam Srinivasan & Rakesh K Mishra
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
######################################################################################################################################
#Usage: perl cdBEST_basic_v1.0.pl
#last modified on 3rd November 2011
#This tool requires perl alone, please see the readme file for more information or contact us by email
#Email: asrini@ccmb.res.in, srinivasan.ccmb@gmail.com, mishra@ccmb.res.in
######################################################################################################################################





use strict;
use warnings;
use Cwd;


        my $usage1 = "\nPlease provide a input fasta file\n\nUsage: perl cdBEST_basic_v1.0.pl input_fasta_file\n";
        
        
        
        #to get the input files
        my $input_fasta_file_name = shift or die $usage1;
        
        my$start_time='';
        my$stop_time='';
        $start_time=localtime(time);
        print "\ncdBEST_basic_v1.0 started at: $start_time\n";
        
        my $detailed_output_file_name = 'motif_and_summary_output.txt';
        my $sequence_out_file_name ='sequence_output.txt';
        my $table_output_file_name = 'table_output.txt';
        my$hits_table_file_name = 'hits_table.txt';
        my$overlapping_unique_hits_file_name ='hits_tables_sorted_and_unique.txt';
        
    
        #pre-set parameters common to all boundary types
        my $window_size = 750;
        my $window_increment = 10;
        
        
        
        
        #pre-set parameters to various boundary types
        my $total_no_motifs_required_fab7_type=8;
        my $total_no_motifs_required_fab8_type=2;
        my $total_no_motifs_required_scs_type=8;
        my $total_no_motifs_required_be28_type=6;
        my $total_no_motifs_required_gypsy_type=2;
        
        
        my $average_gap_set_fab7_type=90;
        my $average_gap_set_fab8_type=90;
        my $average_gap_set_scs_type=90;
        my $average_gap_set_be28_type=90;
        my $average_gap_set_gypsy_type=125;
        
        
        my $score_set_fab7_type=60;
        my $score_set_fab8_type=75;
        my $score_set_scs_type=37.07;
        my $score_set_be28_type=42.54;
        my $score_set_gypsy_type=313.2;
        
        
        
        my $type_specific_criteria_fab7_type=2;
        my $type_specific_criteria_fab8_type=1;
        
        my $type_specific_criteria_scs_type=5;
        my $type_specific_criteria_1_scs_type=2;
        
        my $type_specific_criteria_be28_type=6;
        my $type_specific_criteria_1_be28_type=2;
        
        my $type_specific_criteria_gypsy_type=2;
        
        
      
        
        
        
        
      
      
        #to initiate undefined varaibles (latter used for comparision between first and second cluster)
        my$first_cluster_start=0;
        my$first_cluster_stop=0;
        my$first_cluster_length=0;
        
        #a flag setting to identify, if the hit is found or not(value of 1 means finding of a hit)
        my$hit_found=0;
        
        
        #initiation for return values from boundary predict subroutine
        my$hit_found_returned=0;
        my$cluster_start_returned=0;
        my$cluster_stop_returned=0;
        my$boundary_id_returned=0;
        my$cluster_length_returned=0;
        
        
        
        
        
        my$current_seq_length_in_full_750bp_window=0;
        
        
       
 
        
        
	
	######################################## Boundary Prediction part starts ########################################################
			
        #Now open the file specified by $input_fasta_file_name
	open (INFILE, $input_fasta_file_name) or die( "Cannot open file for input: $!" );
	$/ = undef;
	my $sequenceEntry = <INFILE>;              
	
        close(INFILE)or die ("Cannot close file for output: $!");
        
        my $sequenceTitle;
        
            if ($sequenceEntry =~ m/^>([^\n\cM]+)/)
				
                {
                    $sequenceTitle = $1;
                    $sequenceEntry =~ s/^>[^\n\cM]+//;
                }
			
	    else
                    
                {
    		    $sequenceTitle = "No_title_was_found";
                    
                }
                        
			
	$sequenceEntry =~ s/[^gatcryswkmbdhvn]//ig;
	my $sequenceLength = length($sequenceEntry);
                        
			
	#Write header information to the file specified by $table_output_file_name.
	open(OUTFILE, ">" . $table_output_file_name) or die ("Cannot open file for output: $!");
	print(OUTFILE "Boundary\tStart-Stop\t\tSize\t\tScore\tDistance (kb)\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
                        
                        
        #Write header information to the file specified.
	open(OUTFILE, ">" . $detailed_output_file_name) or die ("Cannot open file for output: $!");
	print(OUTFILE "**********************************************************************************\n");
	print(OUTFILE "********Result of chromatin domain Boundary Element Search Tool (cdBEST) *********\n");  
	print(OUTFILE "***********************AUTHORS: A.SRINIVASAN & RAKESH K MISHRA********************\n\n");
	print(OUTFILE "INPUT FILE USED: $input_fasta_file_name \n");
	print(OUTFILE "SEQUENCE TITLE: $sequenceTitle, SEQUENCE LENGTH: $sequenceLength.\n\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
                        
                        
	#to create sequence ouput file		
	open(OUTFILE, ">" . $sequence_out_file_name) or die ("Cannot open file for output: $!");
	close(OUTFILE) or die ("Cannot close file for output: $!");
			
	#to create all_hits_table.txt file which will be used by sort_hits.pl	
	open(OUTFILE, ">" . $hits_table_file_name) or die ("Cannot open file for output: $!");
	print(OUTFILE "##Name\t\tID-No\tStart\tStop\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
 	
        
        ###Start of prediction process
	my$boundary_id_start=0;
	#default start of sequence
        my$start=0;    
        #default value for end of sequence
        my$end=0;
       
        print"\nSearching for boundaries..\n";
        
        
        #boundary search or prediction is performed while the start sequence is less than or equal to sequence length               
        while ($start<=$sequenceLength)
        {
            
			#end of a given window calculated by adding the window size
			$end=$start+$window_size;
                        
                        
                        
			#to limit the end sequence to the maximum seq length
			if($end>=$sequenceLength)
			{$end=$sequenceLength;}
                        
                        my$current_window_sequence='';
                        
			#to get current seq
			$current_window_sequence = substr($sequenceEntry, $start, $end - $start);
			
                        
                        my$current_seq_length=0;
                        $current_seq_length = length($current_window_sequence);
			$current_seq_length_in_full_750bp_window=$current_seq_length;
                        
                        
                        
                        my$transition_sequence='';
                        
                        
                        
                        #to assigin the $current_window_sequence to $transition_sequence
                        $transition_sequence=$current_window_sequence;
                       
                        
                        
                        #complement the sequence
                        $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                        
                        #reverse the sequence to get the reverse complement
                        
                        my$current_window_reverse_complement_sequence='';
                        $current_window_reverse_complement_sequence=reverse($transition_sequence);
                        
                        
                        
                        
                        #750 window call of boundary_predict sub
                        ######################################################
                        ######################################################
                        ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                              $total_no_motifs_required_fab7_type,
                                                                                                                                                              $total_no_motifs_required_fab8_type,
                                                                                                                                                              $total_no_motifs_required_scs_type,
                                                                                                                                                              $total_no_motifs_required_be28_type,
                                                                                                                                                              $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                              $average_gap_set_fab7_type,
                                                                                                                                                              $average_gap_set_fab8_type,
                                                                                                                                                              $average_gap_set_scs_type,
                                                                                                                                                              $average_gap_set_be28_type,
                                                                                                                                                              $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                              $score_set_fab7_type,
                                                                                                                                                              $score_set_fab8_type,
                                                                                                                                                              $score_set_scs_type,
                                                                                                                                                              $score_set_be28_type,
                                                                                                                                                              $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                              $type_specific_criteria_fab7_type,
                                                                                                                                                              $type_specific_criteria_fab8_type,
                                                                                                                                                              $type_specific_criteria_scs_type,
                                                                                                                                                              $type_specific_criteria_1_scs_type,
                                                                                                                                                              $type_specific_criteria_be28_type,
                                                                                                                                                              $type_specific_criteria_1_be28_type,
                                                                                                                                                              $type_specific_criteria_gypsy_type,
                                                                                                                                                              
                                                                                                                                                              
                                                                                                                                                              
                                                                                                                                                              $sequenceTitle,
                                                                                                                                                              $start,
                                                                                                                                                              $end,
                                                                                                                                                              $current_window_sequence,
                                                                                                                                                              $current_window_reverse_complement_sequence,
                                                                                                                                                              $current_seq_length,
                                                                                                                                                              $first_cluster_start,
                                                                                                                                                              $first_cluster_stop,
                                                                                                                                                              $boundary_id_start,
                                                                                                                                                              $first_cluster_length,
                                                                                                                                                              
                                                                                                                                                              
                                                                                                                                                              );
                        
                        #to check if the hit is found or not
                        $hit_found=$hit_found_returned;
                        
                        if($hit_found==1){
                                             
                                            $first_cluster_start=$cluster_start_returned;
                                            $first_cluster_stop=$cluster_stop_returned;			
                                            $boundary_id_start=$boundary_id_returned;
                                            $first_cluster_length=$cluster_length_returned;
                                          
                                          }
                                
                                
                                
                            #start of 500A sub window
                            if ($hit_found==0)
                        
                            {
                                
                                #to get the end postion of first 500bp out of 750 bp window, here end is start+500, start remains the same
                                my$end_500A=$start+500;
                                
                                #to limit the end sequence to the maximum seq length
                                if($end_500A>=$sequenceLength)
                                {$end_500A=$sequenceLength;}
                                
                                
                                #to get current seq
                                $current_window_sequence = substr($sequenceEntry, $start, $end_500A - $start);
                                $current_seq_length = length($current_window_sequence);
                                
                                
                                #to get the reverse complement of current seq
                                #to assigin the $current_window_sequence to $transition_sequence
                                $transition_sequence=$current_window_sequence;
                                
                                
                                
                                #complement the sequence
                                $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                                
                                
                                
                                #reverse the sequence to get the reverse complement
                                $current_window_reverse_complement_sequence=reverse($transition_sequence);
                                
                                
                                #500A window call of boundary_predict sub
                                ######################################################
                                ######################################################
                                ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                                        $total_no_motifs_required_fab7_type,
                                                                                                                                                                        $total_no_motifs_required_fab8_type,
                                                                                                                                                                        $total_no_motifs_required_scs_type,
                                                                                                                                                                        $total_no_motifs_required_be28_type,
                                                                                                                                                                        $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                        $average_gap_set_fab7_type,
                                                                                                                                                                        $average_gap_set_fab8_type,
                                                                                                                                                                        $average_gap_set_scs_type,
                                                                                                                                                                        $average_gap_set_be28_type,
                                                                                                                                                                        $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                        $score_set_fab7_type,
                                                                                                                                                                        $score_set_fab8_type,
                                                                                                                                                                        $score_set_scs_type,
                                                                                                                                                                        $score_set_be28_type,
                                                                                                                                                                        $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                        $type_specific_criteria_fab7_type,
                                                                                                                                                                        $type_specific_criteria_fab8_type,
                                                                                                                                                                        $type_specific_criteria_scs_type,
                                                                                                                                                                        $type_specific_criteria_1_scs_type,
                                                                                                                                                                        $type_specific_criteria_be28_type,
                                                                                                                                                                        $type_specific_criteria_1_be28_type,
                                                                                                                                                                        $type_specific_criteria_gypsy_type,
                                                                                                                                                                       
                                                                                                                                                                       
                                                                                                                                                                        $sequenceTitle,
                                                                                                                                                                        $start,
                                                                                                                                                                        $end_500A,
                                                                                                                                                                        $current_window_sequence,
                                                                                                                                                                        $current_window_reverse_complement_sequence,
                                                                                                                                                                        $current_seq_length,
                                                                                                                                                                        $first_cluster_start,
                                                                                                                                                                        $first_cluster_stop,
                                                                                                                                                                        $boundary_id_start,
                                                                                                                                                                        $first_cluster_length 
                                                                                                                                                                      );
                                
                                
                                $hit_found=$hit_found_returned;
                                
                                        if($hit_found==1)
                                        
                                            {
                                            
                                            $first_cluster_start=$cluster_start_returned;
                                            $first_cluster_stop=$cluster_stop_returned;			
                                            $boundary_id_start=$boundary_id_returned;
                                            $first_cluster_length=$cluster_length_returned;
                                           
                                            }
                                
                            }#end of 500A sub window
           
           
           
           
           
           
            
                                    #start of 500B sub window
                                    #this part is evaluvated if 500A fails to detect any boudary, start is changed as $start+250, end is same as in 750 bp
                                    if (($hit_found==0) && ($current_seq_length_in_full_750bp_window>=500))
                                        
                                    {
                                        #window size is 5000bp so the end is start + 1000
                                        my$start_500B=$start+250;
                                        #to limit the end sequence to the maximum seq length
                                        
                                        
                                        if($end>=$sequenceLength)
                                        {$end=$sequenceLength;}
                                        
                                        
                                        #to get current seq
                                        $current_window_sequence = substr($sequenceEntry, $start_500B, $end - $start_500B);
                                        $current_seq_length = length($current_window_sequence);
                                        
                                        
                                        #to get the reverse complement of current seq
                                        #to assigin the $current_window_sequence to $transition_sequence
                                        $transition_sequence=$current_window_sequence;
                                        
                                        
                                        #complement the sequence
                                        $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                                        
                                        
                                        #reverse the sequence to get the reverse complement
                                        $current_window_reverse_complement_sequence=reverse($transition_sequence);
                                        
                                        
                                        
                                        #500B window call of boundary_predict sub
                                        ######################################################
                                        ######################################################
                                        ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                                                $total_no_motifs_required_fab7_type,
                                                                                                                                                                                $total_no_motifs_required_fab8_type,
                                                                                                                                                                                $total_no_motifs_required_scs_type,
                                                                                                                                                                                $total_no_motifs_required_be28_type,
                                                                                                                                                                                $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                $average_gap_set_fab7_type,
                                                                                                                                                                                $average_gap_set_fab8_type,
                                                                                                                                                                                $average_gap_set_scs_type,
                                                                                                                                                                                $average_gap_set_be28_type,
                                                                                                                                                                                $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                $score_set_fab7_type,
                                                                                                                                                                                $score_set_fab8_type,
                                                                                                                                                                                $score_set_scs_type,
                                                                                                                                                                                $score_set_be28_type,
                                                                                                                                                                                $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                $type_specific_criteria_fab7_type,
                                                                                                                                                                                $type_specific_criteria_fab8_type,
                                                                                                                                                                                $type_specific_criteria_scs_type,
                                                                                                                                                                                $type_specific_criteria_1_scs_type,
                                                                                                                                                                                $type_specific_criteria_be28_type,
                                                                                                                                                                                $type_specific_criteria_1_be28_type,
                                                                                                                                                                                $type_specific_criteria_gypsy_type,
                                                                                                                                                                       
                                                                                                                                                                                $sequenceTitle,
                                                                                                                                                                                $start_500B,
                                                                                                                                                                                $end,
                                                                                                                                                                                $current_window_sequence,
                                                                                                                                                                                $current_window_reverse_complement_sequence,
                                                                                                                                                                                $current_seq_length,
                                                                                                                                                                                $first_cluster_start,
                                                                                                                                                                                $first_cluster_stop,
                                                                                                                                                                                $boundary_id_start,
                                                                                                                                                                                $first_cluster_length
                                                                                                                                                                             );
                                        
                                        
                                            $hit_found=$hit_found_returned;
                                                
                                            if($hit_found==1)
                                                
                                            {
                                                
                                                    
                                                    $first_cluster_start=$cluster_start_returned;
                                                    $first_cluster_stop=$cluster_stop_returned;			
                                                    $boundary_id_start=$boundary_id_returned;
                                                    $first_cluster_length=$cluster_length_returned;
                                                    
                                            }
                                        
                                        
                                    }#end of 500B sub window	
           
            
            
            
            
            
            
                                            #start of 250A sub window
                                            #this part is evaluvated if 750, 500A and 500B fails to detect any boudary, $start is same as 750, end is modified as $start+250
                                            if ($hit_found==0) 
                                                
                                            {
                                                
                                                
                                                #to get the end postion of first 500bp out of 750 bp window, start remains same
                                                my$end_250A=$start+250;
                                                #to limit the end sequence to the maximum seq length
                                                    
                                                
                                                if($end_250A>=$sequenceLength)
                                                {$end_250A=$sequenceLength;}
                                                
                                                
                                                #to get current seq
                                                $current_window_sequence = substr($sequenceEntry, $start, $end_250A - $start);
                                                $current_seq_length = length($current_window_sequence);
                                                
                                                
                                                
                                                #to get the reverse complement of current seq
                                                #to assigin the $current_window_sequence to $transition_sequence
                                                $transition_sequence=$current_window_sequence;
                                                
                                                
                                                #complement the sequence
                                                $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                                                
                                                #reverse the sequence to get the reverse complement
                                                $current_window_reverse_complement_sequence=reverse($transition_sequence);
                                                    
                                                    
                                                    
                                                #250A window call of boundary_predict sub
                                                ######################################################
                                                ######################################################
                                                ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                                                        $total_no_motifs_required_fab7_type,
                                                                                                                                                                                        $total_no_motifs_required_fab8_type,
                                                                                                                                                                                        $total_no_motifs_required_scs_type,
                                                                                                                                                                                        $total_no_motifs_required_be28_type,
                                                                                                                                                                                        $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                        $average_gap_set_fab7_type,
                                                                                                                                                                                        $average_gap_set_fab8_type,
                                                                                                                                                                                        $average_gap_set_scs_type,
                                                                                                                                                                                        $average_gap_set_be28_type,
                                                                                                                                                                                        $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                        $score_set_fab7_type,
                                                                                                                                                                                        $score_set_fab8_type,
                                                                                                                                                                                        $score_set_scs_type,
                                                                                                                                                                                        $score_set_be28_type,
                                                                                                                                                                                        $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                        $type_specific_criteria_fab7_type,
                                                                                                                                                                                        $type_specific_criteria_fab8_type,
                                                                                                                                                                                        $type_specific_criteria_scs_type,
                                                                                                                                                                                        $type_specific_criteria_1_scs_type,
                                                                                                                                                                                        $type_specific_criteria_be28_type,
                                                                                                                                                                                        $type_specific_criteria_1_be28_type,
                                                                                                                                                                                        $type_specific_criteria_gypsy_type,
                                                                                                                                                                       
                                                                                                                                                                                        $sequenceTitle,
                                                                                                                                                                                        $start,
                                                                                                                                                                                        $end_250A,
                                                                                                                                                                                        $current_window_sequence,
                                                                                                                                                                                        $current_window_reverse_complement_sequence,
                                                                                                                                                                                        $current_seq_length,
                                                                                                                                                                                        $first_cluster_start,
                                                                                                                                                                                        $first_cluster_stop,
                                                                                                                                                                                        $boundary_id_start,
                                                                                                                                                                                        $first_cluster_length
                                                                                                                                                                                     );
                                                    
                                                    
                                                    $hit_found=$hit_found_returned;
                                                    
                                                    if($hit_found==1)
                                                    
                                                        {
                                                            
                                                            
                                                            $first_cluster_start=$cluster_start_returned;
                                                            $first_cluster_stop=$cluster_stop_returned;			
                                                            $boundary_id_start=$boundary_id_returned;
                                                            $first_cluster_length=$cluster_length_returned;
                                                            
                                                            
                                                        }
                                                        
                                                        
                                            }#end of 250A sub window	
                            
                            
                            
                            
                            
                                                    #start of 250B sub window
                                                    #this part is evaluvated if 750, 500A, 500B and 250A fails to detect any boudary, start is changed as $start+250, end is $start+500
                                                    if (($hit_found==0) && ($current_seq_length_in_full_750bp_window>=250))
                                                    
                                                    {
                                                        #to get the start postion of second 250bp out of 750 bp window, start=start+250, the end is start+500
                                                        my$start_250B=$start+250;
                                                        
                                                        #to limit the end sequence to the maximum seq length
                                                        
                                                        my$end_250B=$start+500;
                                                        
                                                        if($end_250B>=$sequenceLength)
                                                        {$end_250B=$sequenceLength;}
                                                        
                                                           
                                                        #to get current seq
                                                        $current_window_sequence = substr($sequenceEntry, $start_250B, $end_250B - $start_250B);
                                                        $current_seq_length = length($current_window_sequence);
                                                        
                                                        
                                                        #to get the reverse complement of current seq
                                                        #to assigin the $current_window_sequence to $transition_sequence
                                                        $transition_sequence=$current_window_sequence;
                                                        
                                                        
                                                        #complement the sequence
                                                        $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                                                        
                                                        
                                                        #reverse the sequence to get the reverse complement
                                                        $current_window_reverse_complement_sequence=reverse($transition_sequence);
                                                          
                                                        
                                                        #250B window call of boundary_predict sub
                                                        ######################################################
                                                        ######################################################    
                                                        ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                                                                $total_no_motifs_required_fab7_type,
                                                                                                                                                                                                $total_no_motifs_required_fab8_type,
                                                                                                                                                                                                $total_no_motifs_required_scs_type,
                                                                                                                                                                                                $total_no_motifs_required_be28_type,
                                                                                                                                                                                                $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                $average_gap_set_fab7_type,
                                                                                                                                                                                                $average_gap_set_fab8_type,
                                                                                                                                                                                                $average_gap_set_scs_type,
                                                                                                                                                                                                $average_gap_set_be28_type,
                                                                                                                                                                                                $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                $score_set_fab7_type,
                                                                                                                                                                                                $score_set_fab8_type,
                                                                                                                                                                                                $score_set_scs_type,
                                                                                                                                                                                                $score_set_be28_type,
                                                                                                                                                                                                $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                $type_specific_criteria_fab7_type,
                                                                                                                                                                                                $type_specific_criteria_fab8_type,
                                                                                                                                                                                                $type_specific_criteria_scs_type,
                                                                                                                                                                                                $type_specific_criteria_1_scs_type,
                                                                                                                                                                                                $type_specific_criteria_be28_type,
                                                                                                                                                                                                $type_specific_criteria_1_be28_type,
                                                                                                                                                                                                $type_specific_criteria_gypsy_type,
                                                                                                                                                                                              
                                                                                                                                                                                                $sequenceTitle,
                                                                                                                                                                                                $start_250B,
                                                                                                                                                                                                $end_250B,
                                                                                                                                                                                                $current_window_sequence,
                                                                                                                                                                                                $current_window_reverse_complement_sequence,
                                                                                                                                                                                                $current_seq_length,
                                                                                                                                                                                                $first_cluster_start,
                                                                                                                                                                                                $first_cluster_stop,
                                                                                                                                                                                                $boundary_id_start,
                                                                                                                                                                                                $first_cluster_length
                                                                                                                                                                                             );
                                                            
                                                            
                                                            $hit_found=$hit_found_returned;
                                                                
                                                                if($hit_found==1)
                                                                
                                                                {
                                                                    
                                                                    
                                                                    $first_cluster_start=$cluster_start_returned;
                                                                    $first_cluster_stop=$cluster_stop_returned;			
                                                                    $boundary_id_start=$boundary_id_returned;
                                                                    $first_cluster_length=$cluster_length_returned;
                                                                  
                                                                    
                                                                }
                                                        
                                                        
                                                    }#end of 250B sub window	                           
                            
                            
                            
                            
                                                            #start of 250C sub window
                                                            #this part is evaluvated if 750, 500A, 500B, 250A and 250B fails to detect any boudary, start is changed as $start+500, end is same as 750
                                                             if (($hit_found==0) && ($current_seq_length_in_full_750bp_window>=500))
                                                               
                                                            {
                                                                
                                                                #to get the start postion of third 250bp out of 750 bp window, start=start+500, the end is same as main window 
                                                                 my$start_250C=$start+500;
                                                                #to limit the end sequence to the maximum seq length
                                                                    
                                                                #to get current seq
                                                                $current_window_sequence = substr($sequenceEntry, $start_250C, $end - $start_250C);
                                                                $current_seq_length = length($current_window_sequence);
                                                                
                                                                #to get the reverse complement of current seq
                                                                #to assigin the $current_window_sequence to $transition_sequence
                                                                $transition_sequence=$current_window_sequence;
                                                                
                                                                
                                                                #complement the sequence
                                                                $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                                                                
                                                                
                                                                #reverse the sequence to get the reverse complement
                                                                $current_window_reverse_complement_sequence=reverse($transition_sequence);
                                                                
                                                                
                                                                
                                                                #250C window call of boundary_predict sub
                                                                ######################################################
                                                                ######################################################    
                                                                ($hit_found_returned,$cluster_start_returned,$cluster_stop_returned,$boundary_id_returned,$cluster_length_returned)=boundary_predict (
                                                                                                                                                                                                        $total_no_motifs_required_fab7_type,
                                                                                                                                                                                                        $total_no_motifs_required_fab8_type,
                                                                                                                                                                                                        $total_no_motifs_required_scs_type,
                                                                                                                                                                                                        $total_no_motifs_required_be28_type,
                                                                                                                                                                                                        $total_no_motifs_required_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                        $average_gap_set_fab7_type,
                                                                                                                                                                                                        $average_gap_set_fab8_type,
                                                                                                                                                                                                        $average_gap_set_scs_type,
                                                                                                                                                                                                        $average_gap_set_be28_type,
                                                                                                                                                                                                        $average_gap_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                        $score_set_fab7_type,
                                                                                                                                                                                                        $score_set_fab8_type,
                                                                                                                                                                                                        $score_set_scs_type,
                                                                                                                                                                                                        $score_set_be28_type,
                                                                                                                                                                                                        $score_set_gypsy_type,
                                                                                                                                                              
                                                                                                                                                                                                        $type_specific_criteria_fab7_type,
                                                                                                                                                                                                        $type_specific_criteria_fab8_type,
                                                                                                                                                                                                        $type_specific_criteria_scs_type,
                                                                                                                                                                                                        $type_specific_criteria_1_scs_type,
                                                                                                                                                                                                        $type_specific_criteria_be28_type,
                                                                                                                                                                                                        $type_specific_criteria_1_be28_type,
                                                                                                                                                                                                        $type_specific_criteria_gypsy_type,
                                                                                                                                                                                 
                                                                                                                                                                                                        $sequenceTitle,
                                                                                                                                                                                                        $start_250C,
                                                                                                                                                                                                        $end,
                                                                                                                                                                                                        $current_window_sequence,
                                                                                                                                                                                                        $current_window_reverse_complement_sequence,
                                                                                                                                                                                                        $current_seq_length,
                                                                                                                                                                                                        $first_cluster_start,
                                                                                                                                                                                                        $first_cluster_stop,
                                                                                                                                                                                                        $boundary_id_start,
                                                                                                                                                                                                        $first_cluster_length
                                                                                                                                                                                                     );
                                                              
                                                                
                                                                $hit_found=$hit_found_returned;
                                                                
                                                                if($hit_found==1)
                                                                
                                                                {
                                                                    
                                                                    
                                                                    $first_cluster_start=$cluster_start_returned;
                                                                    $first_cluster_stop=$cluster_stop_returned;			
                                                                    $boundary_id_start=$boundary_id_returned;
                                                                    $first_cluster_length=$cluster_length_returned;
                                                                    
                                                                    
                                                                }
                                                                
                                                                
                                                            }#end of 250C sub window	                    
                                
                                
        
                
                
                
                
                
                
            # to move the window sequence using window increment value
            $start=$start+$window_increment;
                
                
                
                
        }#end of while loop that performs prediction to the given sequence			      
			      
			      
            
            
        #clear the sequence from the memory   
        #$sequenceEntry='';    
            
            
            
            
            
        
        
        
        # if no hit is found
        #######################################################
        #######################################################
        #######################################################
        if ($boundary_id_start==0)
        {
            
            print"\nNo Boundary hit found for the given input sequence.\n";
                           
        }
        
        
        # if a or more hit is found, then proces the hits and create output files
        #######################################################
        #######################################################
        #######################################################
        else
        {
            # to sort the all_hits_table_file
            #################################
            #################################
            my$sorting_input_file=$hits_table_file_name;
            #my $cmd1 ="perl cdBEST-A.pl \"$sorting_input_file\" ";
            #system($cmd1);
            
            
            
            
            
            
            
            
            
            
            #call a sub to process overlaping hits table and get an array of merged boundaries as return
            ##################################################
            ##################################################
            ##################################################
            my @received_merged_boundaries=();
            @received_merged_boundaries=sort_hitstable_remove_duplicates_find_distance_and_merge_overlapping_hits($sorting_input_file);
            #print"\nYYYYYYYYYYY@received_merged_boundaries\n";
            #end of call for sub 
            ##################################################
            ##################################################
            ##################################################
            
            
            
            
            
            
            
            
            
            #call a sub to repredict_boundaries_for_merged_regions
            ##################################################
            ##################################################
            ##################################################
            re_predict_boundaries_for_merged_regions(@received_merged_boundaries);
            #end of call for sub repredict_boundaries_for_merged_regions
            ##################################################
            ##################################################
            ##################################################
            
            
            

            
            
            
            
            
            
            
            
               
        }# end of if hit is found loop
        



##################################################################### subroutines ###################################################################
#####################################################################################################################################################
#####################################################################################################################################################


#sub to predict boundary in given sub sequence
##############################################
##############################################
sub boundary_predict  

    {
        
                
        my$total_no_motifs_required_fab7_type_value=shift;
        my$total_no_motifs_required_fab8_type_value=shift;
        my$total_no_motifs_required_scs_type_value=shift;
        my$total_no_motifs_required_be28_type_value=shift;
        my$total_no_motifs_required_gypsy_type_value=shift;
                                                                                                                                                              
        my$average_gap_set_fab7_type_value=shift;
        my$average_gap_set_fab8_type_value=shift;
        my$average_gap_set_scs_type_value=shift;
        my$average_gap_set_be28_type_value=shift;
        my$average_gap_set_gypsy_type_value=shift;
    
        my$score_set_fab7_type_value=shift;
        my$score_set_fab8_type_value=shift;
        my$score_set_scs_type_value=shift;
        my$score_set_be28_type_value=shift;
        my$score_set_gypsy_type_value=shift;
        
        my$type_specific_criteria_fab7_type_value=shift;
        my$type_specific_criteria_fab8_type_value=shift;
        my$type_specific_criteria_scs_type_value=shift;
        my$type_specific_criteria_1_scs_type_value=shift;
        my$type_specific_criteria_be28_type_value=shift;
        my$type_specific_criteria_1_be28_type_value=shift;
        my$type_specific_criteria_gypsy_type_value=shift;
        

        my$sequenceTitle_value=shift;
        my$start_value=shift;
        my$end_value=shift;
        my$prediction_sequence=shift;
        my$prediction_sequence_reverse_complement=shift;
        my$prediction_sequence_length=shift;
        my$first_cluster_start_value=shift;
        my$first_cluster_stop_value=shift;
        my$boundary_id_value=shift;
        my$first_cluster_length_value=shift;
        
        
  
        
        my$hit_found=0;
        
        
        
        #initialization of variables used within this function
        my$calculated_average_gap='';
        my$dif_kind_motif=0;
        my$first_motif_start_line='';
        my$second_motif_start_line='';
        my$motif_name1='';
        my$motif_name2='';
        my$motif_gap='';
        my$cluster_start=0;
        my$cluster_stop=0;
        my$cluster_length=0;
                        
                        
                    my$ctcf_found=0;
                    
                    #for CTCF-1
		    my$ctcf_found_1=0;
		    
                    #for CTCF-2
		    my$ctcf_found_2=0;
                    
                    #for CTCF-3
		    my$ctcf_found_3=0;
                    my$ctcf_overlapping_site=0;
                    
                    
		    #for GAGA 
		    my$gaga_found=0;
                    
		    #for Zw5
		    my$Zw5_found=0;
                    
		    #for BEAF
		    my$BEAF_found=0;
                   
		    #for Su_HW1
		    my$SuHWa_found=0;
		    
		    #for Su_HW2
		    my$SuHWb_found=0;
		    
                    my$SuHWb_overlapping_site=0;	
                    my$total_SuHw_found=0;
                    
                    #for Fab-7 motif
		    my$Fab7_motif_found=0;
                    
		    #for Elba
		    my$Elba_found=0;
		    
		    
                    #for total number of motifs with out overlapping Su(Hw) sites
		    my$total_number_of_motifs_after_removing_overlapping=0;
			
		    #An array to hold results
		    my @resultsArray=();
		    my@startArray=();
                    my@stopArray=();
                    
                    
                    
                    
                    
		    #The pattern to match											       #Y  R    Y TGCATA Y   Y   Y    # Y  W  GC M  TACTT H     Y
		    my @boundary_motifs = ("[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]","[TC]AG[AG][TG]G[GT]CGC","[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]","GCTG[AC]G","GAGAG","CGATA","[TC][AG][TC]TGCATA[TC][TC][TC]","[TC][TA]GC[AC]TACTT[TAC][TC]","CCAATTGG","[CA]CAATAAG");
					#motifs(CTCF,Zw5,GAGA,BEAF,SuHw[1],SuHw[2],Fab-7 motif)
                        
                    foreach my$boundary_motifs(@boundary_motifs) 
                        
                        
                        
                    {
                        
                        #Removes the first element in the array and assign it to pattern variable
                        my $pattern = $boundary_motifs;
                        
                        my $strandSearched;
			my $matchingSequence;
			my $firstMatchingPosition;
			my $lastMatchingPosition;
                        my $firstMatchingPosition1;
			my $lastMatchingPosition1;
                        my $firstMatchingPosition2;
			my $lastMatchingPosition2;
                        
                        my$motif_name;
			
                        
                        #Perform the search for the direct strand first.
                        
			$strandSearched = "+";
                        
			while ($prediction_sequence =~ m/($pattern)/gi) 
			
			{
			
    			    $matchingSequence = $1;
    			    $lastMatchingPosition = pos($prediction_sequence);
    			    
			    $firstMatchingPosition = $lastMatchingPosition - length($matchingSequence)+ 1;
                            
                            # count for no motifs
                            
			    if ($pattern eq "[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]")
    			    {$ctcf_found_1=$ctcf_found_1+1;$motif_name="CTCF-1";}
			    
			    
                            elsif ($pattern eq "[TC]AG[AG][TG]G[GT]CGC")
    			    {$ctcf_found_2=$ctcf_found_2+1;$motif_name="CTCF-2";}
                            
                            elsif ($pattern eq "[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]")
    			    {$ctcf_found_3=$ctcf_found_3+1;$motif_name="CTCF-3";}
                            
                            elsif ($pattern eq "GCTG[AC]G")
    			    {$Zw5_found=$Zw5_found+1;$motif_name="Zw5";}
                            
			    elsif ($pattern eq "GAGAG")
    			    {$gaga_found=$gaga_found+1;$motif_name="GAGA";}
                            
			    elsif ($pattern eq "CGATA")
    			    {$BEAF_found=$BEAF_found+1;$motif_name="BEAF";}
                            
			    elsif ($pattern eq "[TC][AG][TC]TGCATA[TC][TC][TC]")
    			    {$SuHWa_found=$SuHWa_found+1;$motif_name="Su(Hw)[1]";}
                            
			    elsif ($pattern eq "[TC][TA]GC[AC]TACTT[TAC][TC]")
    			    {$SuHWb_found=$SuHWb_found+1;$motif_name="Su(Hw)[2]";}
			    
			    elsif ($pattern eq "CCAATTGG")
    			    {$Fab7_motif_found=$Fab7_motif_found+1;$motif_name="F7M";}
			    
			    elsif ($pattern eq "[CA]CAATAAG")
    			    {$Elba_found=$Elba_found+1;$motif_name="Elba";}			
                            
			    $firstMatchingPosition1=$firstMatchingPosition+$start_value;
			    $lastMatchingPosition1=$lastMatchingPosition+$start_value;
                            
                            
			    push( @resultsArray, "$motif_name\t\t$firstMatchingPosition1\t$lastMatchingPosition1\t$strandSearched\t$matchingSequence\t\t");   
                            
                            push( @startArray, "$firstMatchingPosition1\t"); 
                            push( @stopArray, "$lastMatchingPosition1\t");
                            
			}#end of while loop for each + strand match
                        
			
			
			# to search in - strand
			$strandSearched = "-";
                            
                            
                            
			while ($prediction_sequence_reverse_complement =~ m/($pattern)/gi) 
                            
			{
                            
			    $matchingSequence = $1;
                            
			    #These calculations make the positions correspond to the direct strand.
    			    $firstMatchingPosition = $prediction_sequence_length - pos($prediction_sequence_reverse_complement) + 1;
    			    $lastMatchingPosition = $firstMatchingPosition + length($matchingSequence)- 1;
                            
                            
                            
			    # count for no motifs
			    if ($pattern eq "[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]")
    			    {$ctcf_found_1=$ctcf_found_1+1;$motif_name="CTCF-1";}
			    
                            elsif ($pattern eq "[TC]AG[AG][TG]G[GT]CGC")
    			    {$ctcf_found_2=$ctcf_found_2+1;$motif_name="CTCF-2";}
                            
                            elsif ($pattern eq "[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]")
    			    {$ctcf_found_3=$ctcf_found_3+1;$motif_name="CTCF-3";}
			    
			    elsif ($pattern eq "GCTG[AC]G")
    			    {$Zw5_found=$Zw5_found+1;$motif_name="Zw5";}
                          
			    elsif ($pattern eq "GAGAG")
    			    {$gaga_found=$gaga_found+1;$motif_name="GAGA";}
                            
			    elsif ($pattern eq "CGATA")
    			    {$BEAF_found=$BEAF_found+1;$motif_name="BEAF";}
                            
			    elsif ($pattern eq "[TC][AG][TC]TGCATA[TC][TC][TC]")
    			    {$SuHWa_found=$SuHWa_found+1;$motif_name="Su(Hw)[1]";}
                            
			    elsif ($pattern eq "[TC][TA]GC[AC]TACTT[TAC][TC]")
    			    {$SuHWb_found=$SuHWb_found+1;$motif_name="Su(Hw)[2]";}
                            
			    elsif ($pattern eq "[CA]CAATAAG")
    			    {$Elba_found=$Elba_found+1;$motif_name="Elba";}				
                            
                            
                            
			    $firstMatchingPosition2=$firstMatchingPosition+$start_value;
			    $lastMatchingPosition2=$lastMatchingPosition+$start_value;
                            
			    
                            
                            # to avoid calling F7M motif twice
                            if ($pattern ne "CCAATTGG")
    			    {
                               
                              
                                push( @resultsArray, "$motif_name\t\t$firstMatchingPosition2\t$lastMatchingPosition2\t$strandSearched\t$matchingSequence\t\t"); 
                              
                              push( @startArray, "$firstMatchingPosition2\t"); 
                              push( @stopArray, "$lastMatchingPosition2\t"); 
                            }
                            
                         
                          
                         
                          
			}#end of while loop for - strand each match
                        
                        
                        
                        
                    }#end of for each motif search loop 			      
                  
                  
                   
		    #Sort the results by start base using a map-sort-map.
		    @resultsArray = map { $_->[1] }
      		    sort { $a->[0] <=> $b->[0]}
      		    map { [ getStartBase($_), $_] }
      		    @resultsArray;
                    
		    #sub used for sorting
                    sub getStartBase 
                  
		    {
                        
                        my $result = shift();
                        $result =~ m/^\S+\t\t(\d+)\t\d+\t[\+\-]\t\S+/;
                        return $1;
			
                    }
		    
                    
                    
                    
                    
                    
                    #Sort the results by start base using a map-sort-map.
		    @startArray = map { $_->[1] }
      		    sort { $a->[0] <=> $b->[0]}
      		    map { [ getStartBase1($_), $_] }
      		    @startArray;
                    
                    
                    #print"@startArray\n";
                    
		    #sub used for sorting
                    sub getStartBase1 
                  
		    {
                        
                        my $result = shift();
                        $result =~ m/^(\d+)\t/;
                        return $1;
			
                    }
                    
                    
                    #Sort the results by start base using a map-sort-map.
		    @stopArray = map { $_->[1] }
      		    sort { $a->[0] <=> $b->[0]}
      		    map { [ getStartBase2($_), $_] }
      		    @stopArray;
                    
                    
		    #sub used for sorting
                    sub getStartBase2 
                  
		    {
                        
                        my $result = shift();
                        $result =~ m/^(\d+)\t/;
                        return $1;
			
                    }
                    
                    
                    
                    
                    
                    
                    
                    
                    
		    my$numberOfMatches=0;
		    # to find no of element in the array
		    $numberOfMatches = scalar(@resultsArray);
		    
                    
                    
                    # proceed further if no of motifs are more than two else return default values
                    if($numberOfMatches>=2)
                    
		    {
			#to count overlapping Su(Hw) sites
			
                        
			# to find the cluster start and end 
			my$cluster_start_line=$resultsArray[0];
			my$cluster_end_line=$resultsArray[-1];
			
                        
			
			if ($cluster_start_line =~ m/^\S+\t\t(\d+)\t\d+\t[\+\-]\t\S+/)
			
			{$cluster_start=$1;}
			
			
			if ($cluster_end_line =~ m/^\S+\t\t\d+\t(\d+)\t[\+\-]\t\S+/)
			
			{$cluster_stop=$1;}
			
			$cluster_length=$cluster_stop-$cluster_start+1;			
			
                        
                        
                        # to find the cluster start and end 
                        my$new_cluster_start=0;
                        my$new_cluster_stop=0;
                        
                        my$new_cluster_start_line=$startArray[0];
                        if ($new_cluster_start_line =~ m/^(\d+)\t/)
                        {$new_cluster_start=$1;}
                        
                        my$new_cluster_end_line=$stopArray[-1];
                        if ($new_cluster_end_line =~ m/^(\d+)\t/)
                        {$new_cluster_stop=$1;}
                        
                        
                        my$new_cluster_length=$new_cluster_stop-$new_cluster_start+1;
                        
                        
                        
                        
                        
                        
                        
                        
                        
                        
                        
                        
                       
			my$potential_BEAF_binding_sites=0;		
			my$potential_Zw5_binding_sites=0;
			# to add the gap detail to the final file 
                        
			my@resultsArray1=();	
                        
			#to push the first element in the array with 0 gap
			push( @resultsArray1, "$cluster_start_line\t0\n"); 
                        
                      
			#to find the gap between each motif and write it to new array 
			#############################################################
                        #############################################################
                        for (my $i = 0 ; $i <= $numberOfMatches-2; $i++ ) 
                            
			{
                            
			    my$current_first_motif_stop=0;
			    my$current_second_motif_start=0;
                            
			    # to find the cluster start and end
			 
			    $first_motif_start_line=$resultsArray[$i];
			    $second_motif_start_line=$resultsArray[$i+1];
                            
                            
                         
                            if ($first_motif_start_line =~ m/^(\S+)\t\t\d+\t(\d+)\t[\+\-]\t\S+/)
                                    
				{
                                    
                                    $motif_name1=$1;
                                    $current_first_motif_stop=$2;
                                    
				}
                                
                              
                                
			    if ($second_motif_start_line =~ m/^(\S+)\t\t(\d+)\t\d+\t[\+\-]\t\S+/)
                                
				{
                                    
                                    $motif_name2=$1;
                                    $current_second_motif_start=$2;
                                    
                                }
                              
                              
                              
                            $motif_gap=$current_second_motif_start-$current_first_motif_stop;
                                    
                                    
                            if(($motif_name1 eq "BEAF")&&($motif_name2 eq "BEAF")&&($motif_gap<=16))
                            {
                                
                                #print"$motif_name1\t$motif_name2\t$motif_gap\n";
                                $potential_BEAF_binding_sites=$potential_BEAF_binding_sites+1;
                                
                            }  
                               
                            elsif(($motif_name1 eq "Zw5")&&($motif_name2 eq "Zw5")&&($motif_gap<=13))
                            {
                             
                                $potential_Zw5_binding_sites=$potential_Zw5_binding_sites+1;
                                
                            }  
                           
                            elsif((($motif_name1 eq "Su(Hw)[1]")&&($motif_name2 eq "Su(Hw)[2]")&&($motif_gap<0))or(($motif_name2 eq "Su(Hw)[1]")&&($motif_name1 eq "Su(Hw)[2]")&&($motif_gap<0))) 
                            {
                                
                                $SuHWb_overlapping_site=$SuHWb_overlapping_site+1;
                               
                            }  
                          
                            elsif((($motif_name1 eq "CTCF-1")&&($motif_name2 eq "CTCF-2")&&($motif_gap<0))or(($motif_name2 eq "CTCF-1")&&($motif_name1 eq "CTCF-2")&&($motif_gap<0))) 
                            {
                                
                                $ctcf_overlapping_site=$ctcf_overlapping_site+1;
                                
                            }
                           
                            elsif((($motif_name1 eq "CTCF-1")&&($motif_name2 eq "CTCF-3")&&($motif_gap<0))or(($motif_name2 eq "CTCF-1")&&($motif_name1 eq "CTCF-3")&&($motif_gap<0))) 
                            {
                                
                                $ctcf_overlapping_site=$ctcf_overlapping_site+1;
                                
                            }
                            
                            elsif((($motif_name1 eq "CTCF-2")&&($motif_name2 eq "CTCF-3")&&($motif_gap<0))or(($motif_name2 eq "CTCF-2")&&($motif_name1 eq "CTCF-3")&&($motif_gap<0))) 
                            {
                                
                                $ctcf_overlapping_site=$ctcf_overlapping_site+1;
                                
                            }
                          
                            #to push the remaining elements in the array with gap
			    push( @resultsArray1, "$second_motif_start_line\t$motif_gap\n"); 
                            
                          
                           
                        }# closing of writing to new array	
                        
                        
			
			
			
                        
                        if(($ctcf_found_1>=1)or ($ctcf_found_2>=1)or($ctcf_found_3>=1))
                        {$dif_kind_motif=$dif_kind_motif+1;}
                        
                        
			if($Zw5_found>=1)
			{$dif_kind_motif=$dif_kind_motif+1;}
			
			
                        
			if($gaga_found>=1)
			{$dif_kind_motif=$dif_kind_motif+1;}
                        
			
			
			if($BEAF_found>=1)
			{$dif_kind_motif=$dif_kind_motif+1;}
			
			
			
			if(($SuHWa_found>=1)or ($SuHWb_found>=1))
			{$dif_kind_motif=$dif_kind_motif+1;}
			
                        
                        $total_SuHw_found=$SuHWa_found+$SuHWb_found-$SuHWb_overlapping_site;
                        
			
                        
			if($Fab7_motif_found>=1)
			{$dif_kind_motif=$dif_kind_motif+1;}
			
			
			if($Elba_found>=1)
			{$dif_kind_motif=$dif_kind_motif+1;}
			
			
			
			# to calculate the score
			my$cluster_score=0;
			
                        $calculated_average_gap=($cluster_length/$numberOfMatches);
                        $calculated_average_gap= sprintf("%.2f", $calculated_average_gap);
                            
                            
                        #intitiation of score values for each motif
                        my$ctcf_score=0;
                        my$gaga_score=0;
                        my$Zw5_score=0;
                        my$BEAF_score=0;
                        my$SuHWa_score=0;
                        my$SuHWb_score=0;
                        my$Fab7_motif_score=0;
                        my$Elba_score=0;
                            
                        #to find total CTCF sites
                        $ctcf_found=$ctcf_found_1+$ctcf_found_2+$ctcf_found_3;
                            
                        #multiplication of scores for each motif and their occurences
                        $ctcf_score=($ctcf_found_1*150.91)+($ctcf_found_2*61.58)+($ctcf_found_3*165.67);   
                        $gaga_score=$gaga_found*3.64;
                        $Zw5_score=$Zw5_found*5.23;
                        $BEAF_score=$BEAF_found*7.09;
                        $SuHWa_score=$SuHWa_found*156.60;
                        $SuHWb_score=($SuHWb_found-$SuHWb_overlapping_site)*156.60;
                        $Fab7_motif_score=$Fab7_motif_found*73.03;
                        $Elba_score=$Elba_found*14.21;
                      
                      
                        # to find the total no of motifs
                        $total_number_of_motifs_after_removing_overlapping=$numberOfMatches-$SuHWb_overlapping_site;			
                        $total_number_of_motifs_after_removing_overlapping=$numberOfMatches-$ctcf_overlapping_site;
                        
                        # the final score
                        $cluster_score=$ctcf_score+$gaga_score+$Zw5_score+$BEAF_score+$Fab7_motif_score+$SuHWa_score+$SuHWb_score+$Elba_score;
                        $cluster_score= sprintf("%.2f", $cluster_score);
                      
                     
                        
			
                        
                            ##################################################################
                            
                            
                            #print results if it is not a duplicate
                            
                            if((($total_number_of_motifs_after_removing_overlapping>=2)&&($first_cluster_start_value!=$cluster_start)&&($first_cluster_length_value!=$cluster_length))or(($first_cluster_stop_value!=$cluster_stop)&&($total_number_of_motifs_after_removing_overlapping>=2)&&($first_cluster_length_value!=$cluster_length)))
                               
                            {
                                
                                
                                #######8 motifs, 2 different kind of motifs (Fab-7 type)
                                ##############################################################
                                ##############################################################                             
				if( $total_number_of_motifs_after_removing_overlapping>=$total_no_motifs_required_fab7_type_value && $calculated_average_gap<=$average_gap_set_fab7_type_value && $cluster_score>=$score_set_fab7_type_value && $dif_kind_motif>=$type_specific_criteria_fab7_type_value )
                                
				{	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                    
                                    
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$cluster_stop\t$cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    #end of printing for a given seq
                                    
                                    
                                    $hit_found=1;
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);
                                    
                                }#close of printing for Fab-7 boundary type 
                               
                              
                                
                              
                               
                                
                                #######6 BEAF sites and total no motifs is 6 and above
                                ######################################################
                                ######################################################    
                                elsif( $total_number_of_motifs_after_removing_overlapping>=$total_no_motifs_required_be28_type_value && $calculated_average_gap<=$average_gap_set_be28_type_value && $cluster_score>=$score_set_be28_type_value && $BEAF_found>=$type_specific_criteria_be28_type_value && $potential_BEAF_binding_sites>=$type_specific_criteria_1_be28_type_value)
                                   
				{	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                    
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$cluster_stop\t$cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    #end of printing for a given seq
                                    
                                    
                                    
                                    $hit_found=1;
                                    
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);
                                    
                                }#close of printing for BEAF boundary type 
                                
                                
                                
                                
                                
                                
                                
                                
                                
                                #1ctcf site prints output only if the cluster contains 2ctcf sites total of 5 motifs
                                #####################################################################################
                                #####################################################################################
			        elsif($total_number_of_motifs_after_removing_overlapping>=$total_no_motifs_required_fab8_type_value && $calculated_average_gap<=$average_gap_set_fab8_type_value &&  $cluster_score>=$score_set_fab8_type_value && $ctcf_found>=$type_specific_criteria_fab8_type_value )   
                                {	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                    
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$new_cluster_stop\t$new_cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    
                                   
                                    
                                    $hit_found=1;
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$new_cluster_stop,$boundary_id_value,$new_cluster_length);
                                    
                                    
                                }#close of printing for CTCF boundary type
                                
                                
                                
                                #1-1-1 ctcf site prints output only if the cluster contains one of each M1 M2 M3 ctcf sites
                                #####################################################################################
                                #####################################################################################   
                                elsif($ctcf_found_1>=1&& $ctcf_found_2>=1 && $ctcf_found_3>=1)
                                
                                {	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                    
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$new_cluster_stop\t$new_cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    
                                   
                                    
                                    $hit_found=1;
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$new_cluster_stop,$boundary_id_value,$new_cluster_length);
                                    
                                    
                                }#close of printing for CTCF boundary type
                                
                                
                                
                                #######Su(Hw)[1] Su(Hw)[2] based boundary #prints output only if the  Su(Hw)[1] or Su(Hw)[2] site is 2 above.
                                #############################################################################################################
                                #############################################################################################################
                                elsif($total_number_of_motifs_after_removing_overlapping>=$total_no_motifs_required_gypsy_type_value && $calculated_average_gap<=$average_gap_set_gypsy_type_value && $cluster_score>=$score_set_gypsy_type_value  && $total_SuHw_found>=$type_specific_criteria_gypsy_type_value)
                                {	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                 
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$cluster_stop\t$cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    
                                    
                                    $hit_found=1;
                                    
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);    
                                    
                                    
                                    
                                }#close of printing for Su(Hw) boundary type   
                                
                                
                                
                                
                                
                                
                                #######zw5 based boundary #prints output only if the site is 2 above.
                                #####################################################################
                                #####################################################################
                                elsif($total_number_of_motifs_after_removing_overlapping>=$total_no_motifs_required_scs_type_value && $calculated_average_gap<=$average_gap_set_scs_type_value && $cluster_score>=$score_set_scs_type_value && $Zw5_found>=$type_specific_criteria_scs_type_value && $potential_Zw5_binding_sites>=$type_specific_criteria_1_scs_type_value) 
                                {	
                                    
                                    $boundary_id_value=$boundary_id_value+1;
                                    
                                    
                                    open(OUTFILE, "+>>" . $hits_table_file_name) or die ("Cannot open file for output: $!");
                                    print(OUTFILE "BOUNDARY\t$boundary_id_value\t$cluster_start\t$cluster_stop\t$cluster_length\t$calculated_average_gap\t$cluster_score\t$total_number_of_motifs_after_removing_overlapping\n");
                                    close(OUTFILE) or die ("Cannot close file for output: $!");
                                    
                                  
                                   
                                    $hit_found=1;
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);   
                                    
                                    
                                }#close of printing for Zw5 boundary type
                                
                                
                                else
                                {
                                    
                                    
                                    #to return the values to main program
                                    return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);
                                    
                                    
                                }#close of else  
                                
                                
                                
                                
                               
                                
                                
                                
                            }#close of duplicate if loop check    
                            
                            
                            
                            
                        #to return the values to main program
                        else{return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);}
                     
                        
                        
                    }
                   
                    #to return the values to main program
                    else{return ($hit_found,$cluster_start,$cluster_stop,$boundary_id_value,$cluster_length);}
     
      
        
    }#end of sub_prection function 			      

#end of sub_prection function 
##############################################
##############################################
##############################################




































#start of sub sort_hitstable_remove_duplicates_find_distance_and_merge_overlapping_hits
##############################################
##############################################
##############################################
sub sort_hitstable_remove_duplicates_find_distance_and_merge_overlapping_hits
    {
       
        my$c_file_name = shift;
        
        #print"$c_file_name\nNNNNNNNNN";
        #to initiate the overlapping_unique_hits.txt output file
        open(OUTFILE, ">" . $overlapping_unique_hits_file_name) or die ("Cannot open file for output: $!");
        print(OUTFILE "Boundary\tStart\tStop (Size)\tDistance (kb)\n");
        close(OUTFILE) or die ("Cannot close file for output: $!");
        

        
        #empty array intiation to hold hits_table_file
        my @ht_file = (  );
        
        
        #empty array intiation to hold sorted hits_table_file
        my @sorted_ht = (  );
        
        
        #variables used for sorting the list
        my$boundary_start= '';
        my$boundary_stop= '';
        my$boundary_size= '';
        my$boundary_avg_gap= '';
        my$boundary_score= '';
        my$total_no_of_motifs= '';
        



        #open(FILE, $c_file_name) || die "Can't open $c_file_name: $!\n";
        #while (<FILE>) {
                 #if($_ =~ m/^BOUNDARY/){print"$_";push(@ht_file, $_)};
                
                
        #}
        # reading file data using an array
        @ht_file = get_file_data1($c_file_name);
        #print"@ht_file\nMMMMMMMMM";
        
        #close(FILE);      
        
        my$scaler_of_ht_file=scalar(@ht_file);

        #print"\n$scaler_of_ht_file\n";print"\n$ht_file[0]\n";
        #print"Array\t@ht_file\n";
        #Starting of sorting hits tabe
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        #sorting the list and store it in a array
        foreach my$line (@ht_file) 
	{
            #print"$line\nMMMMMMMMM";
            while($line =~ m/BOUNDARY\t\d+\t(\d+)\t(\d+)\t(\d+)\t(\d+\.?\d+)\t(\d+\.?\d+)\t(\d+)/gi)
            {
    		$boundary_start =$1;
		$boundary_stop =$2;
		$boundary_size =$3;
		$boundary_avg_gap=$4;
		$boundary_score=$5;
		$total_no_of_motifs=$6;
		
		#print"$line\nMMMMMMMMM";
		#assign to an array
		push( @sorted_ht, "$boundary_start\t$boundary_stop\t$boundary_size\t$boundary_avg_gap\t$boundary_score\t$total_no_of_motifs\n");					
                
                #sort the array by start base using a map-sort-map.
                @sorted_ht= map { $_->[1] }
                sort { $a->[0] <=> $b->[0]}
                map { [ getStartBase3($_), $_] }
                @sorted_ht;
                
                sub getStartBase3 
                {
                    my $result = shift();
                    $result =~ m/^(\d+)\t\d+\t\d+\t\d+\.?\d+\t\d+\.?\d+\t\d+/;
                    return $1;
                }
                
            }#end of if loop
            
            
            
        }#end of foreach loop
        #End of sorting hits tabe
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        
        
        
        
        
        #Starting of duplicate removal
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        #variables used for duplicate removal
        my$boundary_start1= '';
        my$boundary_stop1= '';
        my$boundary_size1= '';
        my$boundary_avg_gap1= '';
        my$boundary_score1= '';
        my$total_no_of_motifs1= '';
        
        my$boundary_start2= '';
        my$boundary_stop2= '';
        
        #initiate a empty array
        my @unique_boundaries = (  );
        
        #assign the first value of duplecate containing list to a variable and push as a first element in unique array to avoid array being empty
        my$first_line=$sorted_ht [0];
        push(@unique_boundaries, $first_line); 
        
        my$number_hits_in_unique_bounadry= 0;                              
        
        
        
        #removal of duplicates from the list
        foreach my$line1 (@sorted_ht) 
	{
            my$flag=1;
            if($line1 =~ /^(\d+)\t(\d+)\t(\d+)\t(\d+\.?\d+)\t(\d+\.?\d+)\t(\d+)/) 
            {
                $boundary_start1 =$1;
		$boundary_stop1 =$2;
                $boundary_size1 =$3;
		$boundary_avg_gap1=$4;
		$boundary_score1=$5;
                $total_no_of_motifs1=$6;
                
                # to find no of element in the unique boundaries array
                $number_hits_in_unique_bounadry = scalar(@unique_boundaries);
                
                #perform a for loop for unique boundaries array elements
                for (my $i = 0 ; $i < $number_hits_in_unique_bounadry; $i++ ) 
		{
                    if ($unique_boundaries[$i] =~ m/^(\d+)\t(\d+)\t\d+\t\d+\.?\d+\t\d+\.?\d+\t\d+/)       
                    {
			$boundary_start2=$1;
			$boundary_stop2=$2;
                    }        
			
                    #compare the values in the non-unique array with unique array values if it matches set the flag value as 0  
                    if(($boundary_start1==$boundary_start2) &&($boundary_stop1==$boundary_stop2))
                    {
			$flag=0;
			
                    }
                    
                }#end of foreach loop
					
					
					
                                        
                                        
                #if it is unique(flag value is 1) then add to unique array else don't add.
                if($flag==1)
                {
                    push(@unique_boundaries, "$boundary_start1\t$boundary_stop1\t$boundary_size1\t$boundary_avg_gap1\t$boundary_score1\t$total_no_of_motifs1\n"); 
		}
		
                
                
            }#end of if loop
            
            
            
        }#end of foreach loop
        #End of duplicate removal
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        
        
        
        
        
        
        
        
        
        
        
        #Starting of finding distance between boundaries
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        ########to find the gap between boundaries
        my$unique_boundary_start= '';
        my$unique_boundary_stop= '';
        my$unique_boundary_size= '';
        my$unique_boundary_avg_gap= '';
        my$unique_boundary_score= '';
        my$unique_boundary_total_no_of_motifs= '';
        my$unique_boundary_id= '';

        my$previous_unique_boundary_stop=0;
        # to find no of element in the array
        $number_hits_in_unique_bounadry = scalar(@unique_boundaries);

        #to find the gap for each motif
        my$gap=0;
        
        for (my$i = 0; $i < $number_hits_in_unique_bounadry;$i++ )
        {
            if ($unique_boundaries[$i] =~ m/^(\d+)\t(\d+)\t(\d+)\t(\d+\.?\d+)\t(\d+\.?\d+)\t(\d+)/)
            {
                $unique_boundary_start=$1;
		$unique_boundary_stop=$2;
		$unique_boundary_size=$3;
		$unique_boundary_avg_gap=$4;
		$unique_boundary_score=$5;
                $unique_boundary_total_no_of_motifs=$6;
                
                #to give boundary id
		$unique_boundary_id=$i+1;
                
                #to find the distance between boundaries
		$gap=($unique_boundary_start-$previous_unique_boundary_stop);
								
		#to print in output file
		open(OUTFILE, "+>>" . $overlapping_unique_hits_file_name) or die ("Cannot open file for output: $!");
                print(OUTFILE "BOUNDARY\t$unique_boundary_id\t$unique_boundary_start\t$unique_boundary_stop\t$gap\n");
		close(OUTFILE) or die ("Cannot close file for output: $!");
								
		
		
            }  
            
            #to assign the current unique boundary stop to previous unique boundary stop
            $previous_unique_boundary_stop=$unique_boundary_stop;
            
        }#end of gap finding
        #End of finding distance between boundaries
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        
        
       
        
        
        
        
        
        
        
    
        
        
        
        
        
        #Starting of merging overlapping boundaries
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        
        #empty array intiation to hold overlapping_boundaries_table_file
        my @ovl_boundaries_file = (  );
        
        @ovl_boundaries_file = get_file_data($overlapping_unique_hits_file_name);
        
	#print"\n#################\n@ovl_boundaries_file\n";
        my $ovl_boundary_start= 0;
        my $ovl_boundary_stop= 0;
        my $ovl_boundary_distance= 1;
	
        my $ovl_boundary_start1= 0;
        my $ovl_boundary_stop1= 0;
	
        #empty array intiation to hold merged boundaries
        my @merged_boundaries = (  );
	
        #merge the overlapping hits and push it to an array called merged boundaries
        foreach my$line (@ovl_boundaries_file) 
        {
            #if($line =~ /^BOUNDARY\t*\d+\t*(\d+)\t*(\d*)\t*(\W\d+)/) 
            while($line =~ m/BOUNDARY\t*\d+\t(\d+)\t(\d+)\t(\-?\d+)/gi) 	
            {
                $ovl_boundary_start =$1;
                $ovl_boundary_stop =$2;
                $ovl_boundary_distance =$3;
                
                
                
                #if gap distance is equal or more than 1 base the call as separate hits
                if($ovl_boundary_distance>=1)
                {
                    #push it to merged array
                    push( @merged_boundaries, "$ovl_boundary_start\t$ovl_boundary_stop\n");					
                    #to assign first the start value so that it can be used in case of overlapping hits
                    $ovl_boundary_start1=$ovl_boundary_start;
                    $ovl_boundary_stop1=$ovl_boundary_stop;
                    
                }  
             
                
                #else  the hits
                else
                {
                    #delete the last element in the merged array which contains correct start and old(false) stop value in case of overlapping hits
                    pop(@merged_boundaries);
                    
                    if($ovl_boundary_stop>$ovl_boundary_stop1)
                    {
                        #replace the old stop value with new correct value
                        $ovl_boundary_stop1=$ovl_boundary_stop;
                    }
                    
                    #push it to merged array
                    push( @merged_boundaries, "$ovl_boundary_start1\t$ovl_boundary_stop1\n");
                    
                }
                
                
                
                
            }#end of if loop
            
            #print"\n#################\n@merged_boundaries\n";
            
        }#end of foreach loop		
        
        #End of merging overlapping boundaries
        #############################################################################################################################
        #############################################################################################################################
        #############################################################################################################################
        
	
        
        
        #to return the merged_boundaries_array to main
	return @merged_boundaries;
        
        
        
        
       
    }#end of sub merge_hits_table
    
#end of sub sort_hitstable_remove_duplicates_find_distance_and_merge_overlapping_hits
##############################################
##############################################
##############################################










#start of sub re_predict_boundaries_for_merged_regions
##############################################
##############################################
##############################################
sub re_predict_boundaries_for_merged_regions
    {
        
        my @hits = @_;
        
       
        
        #Write header information to the file specified.
        open(OUTFILE, ">" . $table_output_file_name) or die ("Cannot open file for output: $!");
        print(OUTFILE "Boundary\tStart-Stop\t\tSize\t\tScore\tDistance (kb)\n");
        close(OUTFILE) or die ("Cannot close file for output: $!");
        
        
        #Write header information to the file specified.
	open(OUTFILE, ">" . $detailed_output_file_name) or die ("Cannot open file for output: $!");
	print(OUTFILE "**********************************************************************************\n");
	print(OUTFILE "********Result of chromatin domain Boundary Element Search Tool (cdBEST) *********\n");  
	print(OUTFILE "***********************AUTHORS: A.SRINIVASAN & RAKESH K MISHRA********************\n\n");
	print(OUTFILE "INPUT FILE USED: $input_fasta_file_name \n");
	print(OUTFILE "SEQUENCE TITLE: $sequenceTitle, SEQUENCE LENGTH: $sequenceLength.\n\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
        
        
	#to create sequence ouput file		
	open(OUTFILE, ">" . $sequence_out_file_name) or die ("Cannot open file for output: $!");
	close(OUTFILE) or die ("Cannot close file for output: $!");
        
        
        
        
        my $line_hits = '';
        my$merged_boundary_start= '';
        my$merged_boundary_stop= '';
        
        my$merged_boundary_start_actual=0;
        my$merged_boundary_stop_actual=0;
        
        my$cluster_start_print=0;
        my$cluster_stop_print=0;
        
        my$gap_to_print=1;
        
        my$extract=0;
        my$current_window_sequence='';
        my$current_seq_length=0;
        my$transition_sequence='';
        
        my$current_window_reverse_complement_sequence='';
        
        my $boundary_id=0;
        
        
        # for each merged hit perform boundary prediction
        foreach $line_hits (@hits) 
	{
            if($line_hits =~ /^(\d+)\t*(\d+)/) 
            {
    		$merged_boundary_start =$1;
		$merged_boundary_stop=$2;
		
                #minus 1 get the actual sequence (as numbers starts with 0 )
                $merged_boundary_start_actual =$merged_boundary_start-1;
		$merged_boundary_stop_actual=$merged_boundary_stop-1;
               
               
               
                $cluster_start_print=$merged_boundary_start;
                $gap_to_print=($cluster_start_print-$cluster_stop_print+1)/1000;
		$gap_to_print= sprintf("%.1f", $gap_to_print);
		
		$cluster_stop_print=$merged_boundary_stop;
		
                
                $extract=$merged_boundary_stop_actual-$merged_boundary_start_actual+1;
                
                #to get current seq
                $current_window_sequence = substr($sequenceEntry, $merged_boundary_start_actual, $extract);
                
                
                $current_seq_length = length($current_window_sequence);
                    
                #to assigin the $current_window_sequence to $transition_sequence
                $transition_sequence=$current_window_sequence;
                
                #complement the sequence
                $transition_sequence=~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                
                #reverse the sequence to get the reverse complement
                $current_window_reverse_complement_sequence=reverse($transition_sequence);
               
                
                $boundary_id=$boundary_id+1;
                
               
                #call of re_predict_function to predict for a given region 
                re_predict ($sequenceTitle,$current_window_sequence,$current_window_reverse_complement_sequence,$current_seq_length,$boundary_id,$merged_boundary_start_actual,$merged_boundary_stop_actual,$gap_to_print,$table_output_file_name,$detailed_output_file_name,$sequence_out_file_name);
                
                
            }#end of if loop
            
            
	}#end of foreach loop
        
        
       	
        #########################to print the conditions used for running the program############################################################
        open(OUTFILE, "+>>" . $detailed_output_file_name) or die ("Cannot open file for output: $!");			      
        print(OUTFILE "\n\n\n");
        print(OUTFILE "************************************************************************************\n\n");
        print(OUTFILE "Input fasta file used: $input_fasta_file_name\n");
        print(OUTFILE "Sequence title used for prediction: $sequenceTitle\n");
        print(OUTFILE "Total no of hits found: $boundary_id\n\n");
        print(OUTFILE "************************************************************************************\n");
        close(OUTFILE) or die ("Cannot close file for output: $!");
        ##########################################################################End of print ##################################################			      
        
        
        print"\nTotal no. of boundaries found: $boundary_id\n\n";
        
        


        $stop_time=localtime(time);
        print "cdBEST finished at the end: $stop_time\n";


    }
#end of sub re_predict_boundaries_for_merged_regions
##############################################
##############################################
##############################################



#this sub will be called by sub re_predict_boundaries_for_merged_regions
#start of sub re_predict
##############################################
##############################################
##############################################
sub re_predict  
    {	
        my$sequenceTitle_value=shift;
        my$prediction_sequence=shift;
        my$prediction_sequence_reverse_complement=shift;
        my$prediction_sequence_length=shift;
        my$boundary_id_value=shift;
        my$start_value=shift;
        my$end_value=shift;
        my$gap_to_print_value=shift;
        my$table_output_file_name=shift;
        my$detailed_output_file_name=shift;
        my$sequence_out_file_name=shift;
        
        my$dif_kind_motif=0;
        my$motif_name1='';
        my$motif_name2='';
        my$motif_gap='';
        my$cluster_start=0;
        my$cluster_stop=0;
        my$cluster_length=0;
        my$average_gap='';
        
	
	#for CTCF
	my$ctcf_found1=0;
	my$ctcf_found2=0;
	my$ctcf_found3=0;
     
	#for CTCF-1
	my$ctcf_found_1=0;
        
        #for CTCF-2
	my$ctcf_found_2=0;
        
        #for CTCF-3
	my$ctcf_found_3=0;
     
     
     
     
     
     
	#for GAGA 
	my$gaga_found1=0;
	my$gaga_found2=0;
	my$gaga_found3=0;
	
	
	#for Zw5
	my$Zw5_found1=0;
	my$Zw5_found2=0;
	my$Zw5_found3=0;
	
	#for BEAF
	my$BEAF_found1=0;
	my$BEAF_found2=0;
	my$BEAF_found3=0;
        
	#for Su_HW
	my$SuHWa_found1=0;
	my$SuHWa_found2=0;
	my$SuHWa_found3=0;
        
	#for Su_HW
	my$SuHWb_found1=0;
	my$SuHWb_found2=0;
	my$SuHWb_found3=0;
	
	
	#for Fab-7 motif
	my$Fab7_motif_found1=0;
	my$Fab7_motif_found2=0;
	my$Fab7_motif_found3=0;
	
	#for Elba
	my$Elba_found1=0;
	my$Elba_found2=0;
	my$Elba_found3=0;
	
        my$motif_name='';
        my$firstMatchingPosition1=0;
        my$lastMatchingPosition1=0;
        my$firstMatchingPosition2=0;
        my$lastMatchingPosition2=0;
        my$numberOfMatches=0;
      
        
       
	
        #for total number of motifs with out overlapping Su(Hw) sites
	my$total_number_of_motifs_after_removing_overlapping=0;
	
	#An array to hold results
	my @resultsArray=();
	
	#The pattern to match											    #Y  R    Y TGCATA Y   Y   Y    # Y  W  GC M  TACTT H     Y
	my @boundary_motifs = ("[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]","[TC]AG[AG][TG]G[GT]CGC","[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]","GCTG[AC]G","GAGAG","CGATA","[TC][AG][TC]TGCATA[TC][TC][TC]","[TC][TA]GC[AC]TACTT[TAC][TC]","CCAATTGG","[CA]CAATAAG");
	#motifs(CTCF,Zw5,GAGA,BEAF,SuHw[1],SuHw[2],Fab-7 motif)
       
	
        
        foreach my$boundary_motifs(@boundary_motifs) 
        {
		
		#Removes the first element in the array and assign it to pattern variable
		my $pattern = $boundary_motifs;
                
		my $strandSearched;
		my $matchingSequence;
		my $firstMatchingPosition;
		my $lastMatchingPosition;
                
		#Perform the search for the direct strand first.
		
		$strandSearched = "+";
                
		while ($prediction_sequence =~ m/($pattern)/gi) 	
		{
			
    			$matchingSequence = $1;
    			$lastMatchingPosition = pos($prediction_sequence);
    			
			$firstMatchingPosition = $lastMatchingPosition - length($matchingSequence)+ 1;
			
			
                        
			if ($pattern eq "[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]")
    			{$ctcf_found_1=$ctcf_found_1+1;$motif_name="CTCF-M1";}
			    
			    
                        elsif ($pattern eq "[TC]AG[AG][TG]G[GT]CGC")
    			{$ctcf_found_2=$ctcf_found_2+1;$motif_name="CTCF-M2";}
                            
                        elsif ($pattern eq "[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]")
    			{$ctcf_found_3=$ctcf_found_3+1;$motif_name="CTCF-M3";}
			
			
			elsif ($pattern eq "GCTG[AC]G")
    			{$Zw5_found1=$Zw5_found1+1;$motif_name="Zw5";}
			
			elsif ($pattern eq "GAGAG")
    			{$gaga_found1=$gaga_found1+1;$motif_name="GAF";}
			
			elsif ($pattern eq "CGATA")
    			{$BEAF_found1=$BEAF_found1+1;$motif_name="BEAF";}
			
			elsif ($pattern eq "[TC][AG][TC]TGCATA[TC][TC][TC]")
    			{$SuHWa_found1=$SuHWa_found1+1;$motif_name="Su(Hw)-M1";}
			
			elsif ($pattern eq "[TC][TA]GC[AC]TACTT[TAC][TC]")
    			{$SuHWb_found1=$SuHWb_found1+1;$motif_name="Su(Hw)-M2";}
			
			elsif ($pattern eq "CCAATTGG")
    			{$Fab7_motif_found1=$Fab7_motif_found1+1;$motif_name="F7M";}
                        
			elsif ($pattern eq "[CA]CAATAAG")
    			{$Elba_found1=$Elba_found1+1;$motif_name="Elba";}			
			
			$firstMatchingPosition1=$firstMatchingPosition+$start_value;
			$lastMatchingPosition1=$lastMatchingPosition+$start_value;
			
			
			
			
			push( @resultsArray, "$motif_name\t\t$firstMatchingPosition1\t$lastMatchingPosition1\t$strandSearched\t$matchingSequence\t\t");   
			
                        
                }#end of while loop for each + strand match
		
		
		
		
		$strandSearched = "-";
		
                
                
		while ($prediction_sequence_reverse_complement =~ m/($pattern)/gi) 
			
		{
   			
                        $matchingSequence = $1;
    			
			#These calculations make the positions correspond to the direct strand.
    			$firstMatchingPosition = $prediction_sequence_length - pos($prediction_sequence_reverse_complement) + 1;
    			$lastMatchingPosition = $firstMatchingPosition + length($matchingSequence)- 1;
    			
			if ($pattern eq "[AC][CTA][AG]G[GA][TG]G[GT]CGC[TC]")
    			{$ctcf_found_1=$ctcf_found_1+1;$motif_name="CTCF-M1";}
		   
			
                        elsif ($pattern eq "[TC]AG[AG][TG]G[GT]CGC")
    			{$ctcf_found_2=$ctcf_found_2+1;$motif_name="CTCF-M2";}
                  
                        elsif ($pattern eq "[AG][GA]CGCC[AC][TC]C[TC][AG][GT][CT]")
			{$ctcf_found_3=$ctcf_found_3+1;$motif_name="CTCF-M3";}
			
			elsif ($pattern eq "GCTG[AC]G")
    			{$Zw5_found2=$Zw5_found2+1;$motif_name="Zw5";}
			
			elsif ($pattern eq "GAGAG")
    			{$gaga_found2=$gaga_found2+1;$motif_name="GAF";}
			
			elsif ($pattern eq "CGATA")
    			{$BEAF_found2=$BEAF_found2+1;$motif_name="BEAF";}
			
			elsif ($pattern eq "[TC][AG][TC]TGCATA[TC][TC][TC]")
    			{$SuHWa_found2=$SuHWa_found2+1;$motif_name="Su(Hw)-M1";}
			
			elsif ($pattern eq "[TC][TA]GC[AC]TACTT[TAC][TC]")
    			{$SuHWb_found2=$SuHWb_found2+1;$motif_name="Su(Hw)-M2";}
			
                        elsif ($pattern eq "[CA]CAATAAG")
    			{$Elba_found2=$Elba_found2+1;$motif_name="Elba";}				
                        
                        
			$firstMatchingPosition2=$firstMatchingPosition+$start_value;
			$lastMatchingPosition2=$lastMatchingPosition+$start_value;
			
			
			
			if ($pattern ne "CCAATTGG")
    			{
				
				push( @resultsArray, "$motif_name\t\t$firstMatchingPosition2\t$lastMatchingPosition2\t$strandSearched\t$matchingSequence\t\t"); 
				
			}
			
			
                        
                }#end of while loop for - strand each match
                
		
                
                
                
	}#end motif search loop 			      
	
        
        
	#Sort the results by start base using a map-sort-map.
	@resultsArray = map { $_->[1] }
      	sort { $a->[0] <=> $b->[0]}
      	map { [ getStartBase6($_), $_] }
      	@resultsArray;
	
	sub getStartBase6 	
	{
                my $result = shift();
                $result =~ m/^\S+\t\t(\d+)\t\d+\t[\+\-]\t\S+/;
                return $1;
	}
	
        
        
	# to find no of element in the array
        $numberOfMatches = scalar(@resultsArray);
	
        #to count overlapping Su(Hw) sites
	my$SuHWb_overlapping_site=0;	
        
	# to find the cluster start and end 
	my$cluster_start_line=$resultsArray[0];
	my$cluster_end_line=$resultsArray[-1];
	
	if ($cluster_start_line =~ m/^\S+\t\t(\d+)\t\d+\t[\+\-]\t\S+/)
	{$cluster_start=$1;}
	
	if ($cluster_end_line =~ m/^\S+\t\t\d+\t(\d+)\t[\+\-]\t\S+/)
	{$cluster_stop=$1;}
	
        $cluster_length=$cluster_stop-$cluster_start+1;			
        
        my$potential_BEAF_binding_sites=0;		
	my$potential_Zw5_binding_sites=0;
        
        # to add the gap detail to the final file 
	my@resultsArray1=();	
	
	#to push the first element in the array with 0 gap
	push( @resultsArray1, "$cluster_start_line\t0\n"); 
	
	#to find the gap for each motif
	for (my $i = 0 ; $i <= $numberOfMatches-2; $i++ ) 
        {
                
		my$current_first_motif_stop=0;
		my$current_second_motif_start=0;
		
		# to find the cluster start and end
		my$first_motif_start_line=$resultsArray[$i];
		my$second_motif_start_line=$resultsArray[$i+1];
		
		
		if ($first_motif_start_line =~ m/^(\S+)\t\t\d+\t(\d+)\t[\+\-]\t\S+/)
		{
                        $motif_name1=$1;
			$current_first_motif_stop=$2;
		}
		
		
		
		if ($second_motif_start_line =~ m/^(\S+)\t\t(\d+)\t\d+\t[\+\-]\t\S+/)
		{
                        $motif_name2=$1;
			$current_second_motif_start=$2;
			
                }
                
                
                $motif_gap=$current_second_motif_start-$current_first_motif_stop;
                
		if(($motif_name1 eq "BEAF")&&($motif_name2 eq "BEAF")&&($motif_gap<40))
		{
			
                        $potential_BEAF_binding_sites=$potential_BEAF_binding_sites+1;
			
                }  
                
                
		elsif(($motif_name1 eq "Zw5")&&($motif_name2 eq "Zw5")&&($motif_gap<40))
		{
                        
                        $potential_Zw5_binding_sites=$potential_Zw5_binding_sites+1;
			
                }  
                
                
                
		elsif(($motif_name1 eq "Su(Hw)-M1")&&($motif_name2 eq "Su(Hw)-M2")&&($motif_gap<0))
		{
			
                        $SuHWb_overlapping_site=$SuHWb_overlapping_site+1;
                        
                }  
                
                
                        
                
                #to push the remaining elements in the array with gap
                push( @resultsArray1, "$second_motif_start_line\t$motif_gap\n"); 
                
        }#close of gap finding for loop	
	
        
        
	$dif_kind_motif=0;
	#To list no of motifs found  
        
	$ctcf_found3=$ctcf_found1+$ctcf_found2;
        
	$Zw5_found3=$Zw5_found1+$Zw5_found2;
	if($Zw5_found3>=1)
	{$dif_kind_motif=$dif_kind_motif+1;}
	
	$gaga_found3=$gaga_found1+$gaga_found2;
	if($gaga_found3>=1)
	{$dif_kind_motif=$dif_kind_motif+1;}
        
	$BEAF_found3=$BEAF_found1+$BEAF_found2;
	if($BEAF_found3>=1)
	{$dif_kind_motif=$dif_kind_motif+1;}
	
	$SuHWa_found3=$SuHWa_found1+$SuHWa_found2;
	$SuHWb_found3=$SuHWb_found1+$SuHWb_found2;
	if(($SuHWa_found3>=1)or ($SuHWb_found3>=1))
	{$dif_kind_motif=$dif_kind_motif+1;}
        
	$Fab7_motif_found3=$Fab7_motif_found1+$Fab7_motif_found2;
	if($Fab7_motif_found3>=1)
	{$dif_kind_motif=$dif_kind_motif+1;}
	
	$Elba_found3=$Elba_found1+$Elba_found2;
	if($Elba_found3>=1)
	{$dif_kind_motif=$dif_kind_motif+1;}
	
	# to calculate the score  
	my$cluster_score=0;
	if($numberOfMatches==0 or $cluster_length==0)
	{
                $average_gap="Not Applicaple";
                $cluster_score=0;
	}
	
	else
	{
                $average_gap=($cluster_length/$numberOfMatches);
                $average_gap= sprintf("%.2f", $average_gap);
        	
                
		my$ctcf_score=0;
		my$gaga_score=0;
		my$Zw5_score=0;
		my$BEAF_score=0;
		my$SuHWa_score=0;
		my$SuHWb_score=0;
		my$Fab7_motif_score=0;
		my$Elba_score=0;
                
		$ctcf_score=($ctcf_found_1*150.91)+($ctcf_found_2*61.58)+($ctcf_found_3*165.67);
	      
		#$ctcf_score=$ctcf_found3*37.20;
		$gaga_score=$gaga_found3*3.64;
		$Zw5_score=$Zw5_found3*5.23;
		$BEAF_score=$BEAF_found3*7.09;
		$SuHWa_score=$SuHWa_found3*156.60;
		$SuHWb_score=($SuHWb_found3-$SuHWb_overlapping_site)*156.60;
		$Fab7_motif_score=$Fab7_motif_found3*73.03;
		$Elba_score=$Elba_found3*14.21;
		
                
		$total_number_of_motifs_after_removing_overlapping=$numberOfMatches-$SuHWb_overlapping_site;			
		
		$cluster_score=$ctcf_score+$gaga_score+$Zw5_score+$BEAF_score+$Fab7_motif_score+$SuHWa_score+$SuHWb_score+$Elba_score;
		
                
		$cluster_score= sprintf("%.2f", $cluster_score);
                
                
        }#end of score calculation loop
        
        
         ###printing output file 
        open(OUTFILE, "+>>" . $detailed_output_file_name) or die ("Cannot open file for output: $!");
        
	print(OUTFILE "******************************** MOTIF & BOUNDARY SUMMARY ************************\n\n");
        print(OUTFILE "BOUNDARY ID: $boundary_id_value\t\t\tSIZE: $cluster_length bp\t\t\tSCORE: $cluster_score\n\n");	
	print(OUTFILE "LOCATION: $cluster_start - $cluster_stop\tTOTAL NO MOTIFS: $numberOfMatches\t\tAVERAGE GAP: $average_gap\n\n");
        print(OUTFILE "CTCF-M1: $ctcf_found_1\tCTCF-M2: $ctcf_found_2\tCTCF-M3: $ctcf_found_3\tBEAF: $BEAF_found3\t\tGAF: $gaga_found3\n\n");
        print(OUTFILE "Zw5: $Zw5_found3\tSu(Hw)-M1: $SuHWa_found3\tSu(Hw)-M2: $SuHWb_found3\tFab-7 motif(F7M): $Fab7_motif_found3\tElba: $Elba_found3\n\n");
        
        
        #print(OUTFILE "SEQUENCE SEGMENT USED: $start_value- $end_value\n\n");

        print(OUTFILE "**********************************************************************************\n");
	print(OUTFILE "Motif\t\tStart\tStop\tStrand\tSequence found\t\tGap\n\n");
        
        foreach (@resultsArray1) 
        {print (OUTFILE $_);}
	
	print(OUTFILE "----------------------------------------------------------------------------------\n\n\n");
        close(OUTFILE) or die ("Cannot close file for output: $!");
	
        #end of printing for detailed output 

          
        ################################################## to print the output to the file1 in consice format
	open(OUTFILE, "+>>" . $table_output_file_name) or die ("Cannot open file for output: $!");
        #print(OUTFILE "$unique_boundary_id\t\t$unique_boundary_start - $unique_boundary_stop\t($unique_boundary_size)\t\t$unique_boundary_avg_gap\t\t$unique_boundary_score\t\t$unique_boundary_total_no_of_motifs\t\t$gap\n");
								
	print(OUTFILE "$boundary_id_value\t\t$cluster_start - $cluster_stop\t\t$cluster_length\t\t$cluster_score\t$gap_to_print_value\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
        #end of printing for a given seq

	

        ##################################################to print the sequence in separate file.
	open(OUTFILE, "+>>" . $sequence_out_file_name) or die ("Cannot open file for output: $!");
	print(OUTFILE ">BOUNDARY ID: $boundary_id_value Location $cluster_start - $cluster_stop ($cluster_length bp) $sequenceTitle_value\n");
        print(OUTFILE "$prediction_sequence\n\n");
	close(OUTFILE) or die ("Cannot close file for output: $!");
        #end of printing for a given seq
   
     
   
 
}	      
#end of sub re_predict
##############################################
##############################################
##############################################	






















#subroutine to get data from a file given its filename
##############################################
##############################################
##############################################
sub get_file_data
    {
        
        my($filename) = @_;
       
        
        # Initialize variables
        my @filedata = (  );
        
        unless( open(GET_FILE_DATA, $filename) )
        {
        print STDERR "Cannot open file \"$filename\"\n\n";
        exit;
        }
        
        @filedata = <GET_FILE_DATA>;
        
        close GET_FILE_DATA;
        
        return @filedata;
    }

##############################################
##############################################
##############################################








#subroutine to get data from a file given its filename
##############################################
##############################################
##############################################
sub get_file_data1
    {
        
        my($filename1) = @_;
       
        
        # Initialize variables
        my @filedata1 = (  );
        
        unless( open(GET_FILE_DATA, $filename1) )
        {
        print STDERR "Cannot open file \"$filename1\"\n\n";
        exit;
        }
        
        @filedata1 = <GET_FILE_DATA>;
        
        close GET_FILE_DATA;
        
        return @filedata1;
    }

##############################################
##############################################
##############################################

